"
 COPYRIGHT (c) 1998 by eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"


"{ Package: 'stx:libwidg2' }"

TextCollector subclass:#TerminalView
	instanceVariableNames:'inStream outStream readerProcess shellPid kbdSequences
		escapeSequenceTree currentSequence kbdMap escapeLeadingChars
		numberOfColumns numberOfLines shellTerminateAction rangeStartLine
		rangeEndLine state savedCursor shellCommand shellDirectory
		filterStream localEcho translateNLToCRNL inputTranslateCRToNL'
	classVariableNames:''
	poolDictionaries:''
	category:'Views-TerminalViews'
!

!TerminalView class methodsFor:'documentation'!

copyright
"
 COPYRIGHT (c) 1998 by eXept Software AG
              All Rights Reserved

 This software is furnished under a license and may be used
 only in accordance with the terms of that license and with the
 inclusion of the above copyright notice.   This software may not
 be provided or otherwise made available to, or used by, any
 other person.  No title to or ownership of the software is
 hereby transferred.
"

!

documentation
"
    I provide terminal functionality, by interpreting data
    arriving from some stream (typically connected to a command interpreter 
    via a pty, or another system via a modem) 
    and sending my keyboard data to it.

    I am abstract providing general functionality
    - concrete terminal characteristics are defined by concrete subclasses 
    (see VT52TerminalView, VT100TerminalView).

    Concrete applications are: 
        consoles (VT100TerminalView),
        telnet-views (see TelnetTool)
        editor wrappers (if you like emacs/vi)

    <<although my class protocol includes common startup protocol
    (to open a terminalView with a shell or on the output of a command),
    I can be used as a widget within an application (Modem software).

    [author:]
        Claus Gittinger

    [instance variables:]
        inStream                stream where keyboard input is
                                sent to (connected to shells or commands input)

        outStream               stream where the output of the
                                shell or command arrives
                                (read here and displayed in the view)

        readerProcess           process which reads commands    
                                output and sends it to the view

    [start with:]
        VT52TerminalView open
        VT100TerminalView open
        VT52TerminalView openShell
        VT100TerminalView openShell
        VT100TerminalView openOnCommand:'ls -l'
        VT100TerminalView openOnCommand:'dir'

    [see also:]
        TelNetTool
"
! !

!TerminalView class methodsFor:'defaults'!

defaultIcon
    ^ Image fromFile:'bitmaps/xpmBitmaps/terminal_images/xterm-icon.xpm'

    "Created: / 4.8.1998 / 17:48:18 / cg"
    "Modified: / 4.8.1998 / 17:48:31 / cg"
!

defaultNumberOfColumns
    ^ 80

    "Created: / 4.8.1998 / 17:48:18 / cg"
    "Modified: / 4.8.1998 / 17:48:31 / cg"
!

defaultNumberOfLines
    ^ 25

    "Created: / 4.8.1998 / 17:48:18 / cg"
    "Modified: / 4.8.1998 / 17:48:31 / cg"
! !

!TerminalView class methodsFor:'opening'!

open
    ^ self openShell

    "
     VT100TerminalView open
    "

    "Created: / 10.6.1998 / 15:47:25 / cg"
    "Modified: / 9.7.1998 / 17:55:37 / cg"
!

openDummy
    "for testing purposes only - opens a dummy tty-view, which simply
     echoes whatever is typed in"

    |in vt52|

    vt52 := self new.
    in := ForwardingStream on:''.

    in fwdStream:vt52.
    vt52 inStream:in.
    vt52 outStream:in.
    vt52 open

    "
     self openDummy
    "
!

openOnCommand:aCommandString
    "start a command on a pseudo-TTY, open a terminalView on it
     (i.e. this is kind of an xterm)"

    ^ self openOnCommand:aCommandString onExit:[]

    "
     VT100TerminalView openOnCommand:'ls -l'
    "

    "Created: / 9.7.1998 / 17:50:53 / cg"
    "Modified: / 9.7.1998 / 17:57:41 / cg"
!

openOnCommand:aCommandString onExit:aBlock
    "start a command on a pseudo-TTY, open a terminalView on its output
     (i.e. this is kind of an xterm).
     When the command finishes, evaluate aBlock."

    |in top scr vt52|

    top := StandardSystemView new.
    scr := ScrollableView for:self in:top.
    scr origin:0.0@0.0 corner:1.0@1.0.
    vt52 := scr scrolledView.

    vt52 startCommand:aCommandString.
    vt52 shellTerminateAction:[
        top label:('finished: "' , aCommandString , '"'). aBlock value
    ].

    top extent:(scr preferredExtent).
    top label:('executing: "' , aCommandString , '"').
    top iconLabel:'command'.
    top icon:(self defaultIcon).
    top open.

    ^ vt52

    "
     VT100TerminalView openOnCommand:'ls -lR'
     VT100TerminalView openOnCommand:'vi /etc/hosts'
    "

    "Created: / 9.7.1998 / 17:54:34 / cg"
    "Modified: / 4.8.1998 / 17:49:02 / cg"
!

openOnInput:inStream output:outStream
    "open a terminalView on the given streams (which are typically some
     kind of socket or pty). 
     Keys pressed are sent to inStream, text appearing
     from outStream is displayed in the terminal view.
     This can be used to implement things like rlogin
     or telnet views (if connected to a modem, a com-program can also be
     implemented this way)."

    |top scr|

    top := StandardSystemView new.
    scr := self openOnInput:inStream output:outStream in:top.

    top extent:(scr preferredExtent).
    top label:'shell'.
    top iconLabel:'shell'.
    top icon:(self defaultIcon).
    top open.
    ^ top

    "Modified: / 5.5.1999 / 17:25:59 / cg"
!

openOnInput:inStream output:outStream in:aView
    "open a terminalView on the given streams (which are typically some
     kind of socket or pty). 
     Keys pressed are sent to inStream, text appearing
     from outStream is displayed in the terminal view.
     This can be used to implement things like rlogin
     or telnet views (if connected to a modem, a com-program can also be
     implemented this way)."

    |scr vt|

    scr := ScrollableView for:self in:aView.
    scr origin:0.0@0.0 corner:1.0@1.0.
    vt := scr scrolledView.

    vt inStream:inStream.
    vt outStream:outStream.
    vt startReaderProcess.

    ^ scr

    "Modified: / 5.5.1999 / 17:25:59 / cg"
!

openShell
    "start a shell on a pseudo-TTY, open a terminalView on it
     (i.e. this is kind of an xterm)"

    ^ self openShellIn:nil

    "
     VT100TerminalView openShell
    "

    "Modified: / 21.7.1998 / 18:24:55 / cg"
!

openShellIn:aDirectory
    "start a shell on a pseudo-TTY, open a terminalView on it
     (i.e. this is kind of an xterm)"

    |in top scr vt lbl|

    top := StandardSystemView new.
    scr := ScrollableView for:self in:top.
    scr origin:0.0@0.0 corner:1.0@1.0.
    vt := scr scrolledView.

    vt startShellIn:aDirectory.
    vt shellTerminateAction:[top destroy].

    lbl := OperatingSystem isUNIXlike ifTrue:['shell'] ifFalse:['dos'].
    top extent:(scr preferredExtent).
    top label:lbl.
    top iconLabel:lbl.
    top icon:(self defaultIcon).
    top open.

    ^ vt

    "
     VT100TerminalView openShellIn:'/etc'
    "

    "Created: / 20.7.1998 / 18:28:15 / cg"
    "Modified: / 5.5.1999 / 17:27:10 / cg"
! !

!TerminalView class methodsFor:'queries'!

isVisualStartable
    "returns whether this application class can be started via #open"

    self == TerminalView ifTrue:[^ false].
    ^ true

    "Created: / 10.6.1998 / 15:48:43 / cg"
! !

!TerminalView methodsFor:'accessing'!

filterStream:something
    "set a filter stream; if not nil, it gets all incoming data via nextPutAll:.
     (added to allow saving incoming data to a file, but can also be used to catch/filter/lookAt
      incoming data by some other program)"

    filterStream := something.

    "Created: / 28.1.2002 / 20:56:04 / micha"
    "Modified: / 28.1.2002 / 20:56:11 / micha"
!

inStream
    "return the value of the instance variable 'inStream' (automatically generated)"

    ^ inStream
!

inStream:something
    "set the value of the instance variable 'inStream' (automatically generated)"

    inStream := something.
!

inputTranslateCRToNL
    ^ inputTranslateCRToNL
!

inputTranslateCRToNL:something
    inputTranslateCRToNL := something.
!

localEcho:aBoolean
    "enable/disable local echo"

    localEcho := aBoolean

    "Created: / 5.5.1999 / 17:53:16 / cg"
!

outStream
    "return the value of the instance variable 'outStream' (automatically generated)"

    ^ outStream
!

outStream:something
    "set the value of the instance variable 'outStream' (automatically generated)"

    outStream := something.
!

readerProcess
    "return the value of the instance variable 'readerProcess' (automatically generated)"

    ^ readerProcess
!

shellTerminateAction:aBlock
    "set the block which is evaluated when the shell terminates.
     Can be used to close down the application in this case."

    shellTerminateAction := aBlock.

    "Created: / 12.6.1998 / 17:02:58 / cg"
!

translateNLToCRNL
    "return the value of the instance variable 'translateNLToCRNL' (automatically generated)"

    ^ translateNLToCRNL

    "Created: / 28.1.2002 / 20:32:10 / micha"
!

translateNLToCRNL:something
    "set the value of the instance variable 'translateNLToCRNL' (automatically generated)"

    translateNLToCRNL := something.

    "Created: / 28.1.2002 / 20:32:10 / micha"
! !

!TerminalView methodsFor:'cursor handling'!

cursorCol:col
    "check of col is a valid cursor position; return a new col-nr if not.
     Here, the linelength is enforced"

    ^ super cursorCol:(col min:numberOfColumns)

    "Modified: / 10.6.1998 / 15:09:34 / cg"
!

cursorDown:n
    cursorLine + n > list size ifTrue:[
        list := list , (Array new:n).
        self textChanged.
    ].
    super cursorDown:n

    "Modified: / 10.6.1998 / 17:18:41 / cg"
    "Created: / 10.6.1998 / 17:18:50 / cg"
!

cursorLine:l col:col
    "check of col is a valid cursor position; return a new col-nr if not.
     Here, the linelength is enforced"

    ^ super cursorLine:l col:(col min:numberOfColumns)

    "Modified: / 10.6.1998 / 15:09:38 / cg"
!

cursorMovementAllowed
    "return true, if the user may move the cursor around
     (via button-click, or cursor-key with selection).
     Here false is returned - the cursor is only moved by
     cursor positioning escape sequences arriving from the
     stream."

    ^ false

    "Created: / 18.6.1998 / 14:12:02 / cg"
!

numberOfTerminalColumns
    ^ numberOfColumns

    "Created: / 5.5.1999 / 11:46:25 / cg"
!

numberOfTerminalLines
    "/ be careful - this is NOT called numberOfLines,
    "/ since that would interfere with numberOfLines as defined
    "/ in ListView ...
    "/ ... one of the bad sides of subclassing

    ^ numberOfLines

    "Created: / 5.5.1999 / 11:46:18 / cg"
    "Modified: / 5.5.1999 / 11:47:24 / cg"
!

restoreCursor
    |l c|

    l := savedCursor y.
    c := savedCursor x.
    self cursorLine:l col:c.

    "Created: / 14.8.1998 / 13:49:24 / cg"
!

saveCursor
    savedCursor := cursorCol @ cursorLine

    "Created: / 14.8.1998 / 13:48:45 / cg"
    "Modified: / 14.8.1998 / 13:49:32 / cg"
!

validateCursorCol:col inLine:line
    "check of col is a valid cursor position; return a new col-nr if not.
     Here, the linelength is enforced"

    ^ col min:numberOfColumns

    "Modified: / 10.6.1998 / 15:09:41 / cg"
! !

!TerminalView methodsFor:'defaults'!

anyKeyCodes
    ^ IdentityDictionary withKeysAndValues:
        #(
             #Escape      '\e'
             #BackSpace   '\b'
             #Return      '\r'
             #Delete      '\0177'
             #Tab         '\t'
         )

    "Created: / 5.5.1999 / 15:00:37 / cg"
! !

!TerminalView methodsFor:'event handling'!

computeNumberOfLinesShown
    |prevNLines prevNCols|

    prevNCols := (innerWidth // font width).
    prevNLines := nFullLinesShown.

    super computeNumberOfLinesShown.

    ((innerWidth // font width) ~~ prevNCols
    or:[prevNLines ~~ nFullLinesShown]) ifTrue:[
        self defineWindowSize.
    ]

    "Created: / 12.6.1998 / 22:34:39 / cg"
    "Modified: / 20.6.1998 / 19:45:28 / cg"
!

contentsChanged
    super contentsChanged.
    self defineWindowSize.

    "Modified: / 11.6.1998 / 22:51:48 / cg"
    "Created: / 5.5.1999 / 16:30:15 / cg"
!

defineWindowSize
    |fd delta prevNumCols prevNumLines|

"/self halt.
"/self realized ifFalse:[
"/    "/ iconfified
"/    ^ self
"/].
    prevNumCols := numberOfColumns.
    prevNumLines := numberOfLines.

    numberOfColumns := (innerWidth // font width).
    delta := numberOfLines - rangeEndLine.
    numberOfLines := nFullLinesShown.

    ((prevNumCols == numberOfColumns)
    and:[prevNumLines == numberOfLines]) ifTrue:[
        ^ self
    ].

    rangeEndLine notNil ifTrue:[
        rangeEndLine := numberOfLines - delta.
    ].

    "/ any idea, how to do this under windows ?

    OperatingSystem isUNIXlike ifTrue:[
        "/
        "/ tell the pty;
        "/ tell the shell;
        "/
        (inStream notNil 
        and:[inStream isExternalStream
        and:[inStream isOpen]]) ifTrue:[
"/        Transcript showCR:'VT100: changed len to ', numberOfLines printString.

            (OperatingSystem 
                setWindowSizeOnFileDescriptor:inStream fileDescriptor
                width:numberOfColumns
                height:numberOfLines) ifFalse:[
"/            Transcript showCR:'VT100: cannot change windowSize'.
            ].

        ].
        shellPid notNil ifTrue:[
            OperatingSystem sendSignal:OperatingSystem sigWINCH to:shellPid
        ]
    ].

    "Created: / 11.6.1998 / 22:51:39 / cg"
    "Modified: / 5.5.1999 / 19:45:09 / cg"
!

keyPress:aKey x:x y:y
    |rest event rawKey seq|

    "/ somewhat complicated, since some characters
    "/ should go untranslated (CTRL-key),
    "/ even if defined as function keys.

    inStream isNil ifTrue:[^ self].

"/ Transcript showCR:'----'; show:'key:' ; showCR:aKey printString.

    aKey isCharacter ifTrue:[
        "/ send it down to inStream ...
        inStream nextPut:aKey.
        localEcho ifTrue:[
            self nextPut:aKey
        ].
        ^ self
    ].

    (#(Control Control_L Control_R
      Shift Shift_L Shift_R
      Alt Alt_L Alt_R
      Cmd Cmd_L Cmd_R
      Meta Meta_L Meta_R) includes:aKey) ifTrue:[
        ^ self
    ].

    "/
    "/ common translations (Tab, Backspace, F-keys etc.)
    "/
    (aKey == #Return
    and:[inputTranslateCRToNL]) ifTrue:[
        seq := '\n'.
    ] ifFalse:[
        seq := kbdSequences at:aKey ifAbsent:nil.
    ].
    seq notNil ifTrue:[
"/ Transcript show:'->' ; showCR:seq storeString.
        seq := seq withEscapes.

        inStream nextPutAll:seq.
        localEcho ifTrue:[
            self nextPutAll:seq
        ].
        ^ self
    ].

    self sensor ctrlDown ifTrue:[
        (aKey startsWith:'Ctrl') ifTrue:[
            rawKey := aKey
        ] ifFalse:[
            "/ already translated - undo it.
        
            event := WindowGroup lastEventQuerySignal query.
            rawKey := event rawKey.
            rawKey isCharacter ifTrue:[
                rawKey := 'Ctrl' , rawKey.
            ]
        ]
    ] ifFalse:[
        rawKey := self keyboardMap keyAtValue:aKey ifAbsent:aKey.
    ].

    "/
    "/ care for function-keys, which are mapped to Ctrl-x;
    "/
"/ Transcript show:'raw ->' ; showCR:rawKey storeString.

    seq := kbdSequences at:rawKey ifAbsent:nil.
    seq notNil ifTrue:[
"/ Transcript show:'->' ; showCR:seq storeString.
        inStream nextPutAll:(seq withEscapes).
        ^ self
    ].

    (rawKey startsWith:'Ctrl') ifTrue:[
        rest := rawKey copyFrom:5.
        rest size == 1 ifTrue:[
            rest := rest at:1.
            (rest >= $a and:[rest <= $z]) ifTrue:[
"/ Transcript show:'->' ; showCR:(Character value:(rest - $a + 1)) storeString.
                inStream nextPut:(Character value:(rest - $a + 1)).
                ^ self
            ].
            (rest >= $A and:[rest <= $Z]) ifTrue:[
"/ Transcript show:'->' ; showCR:(Character value:(rest - $a + 1)) storeString.
                inStream nextPut:(Character value:(rest - $A + 1)).
                ^ self
            ].
        ]
    ].

    (rawKey startsWith:'Control') ifTrue:[
        ^ self
    ].
    (rawKey startsWith:'Shift') ifTrue:[
        ^ self
    ].
    (rawKey startsWith:'Alt') ifTrue:[
        ^ self
    ].
    (rawKey startsWith:'Cmd') ifTrue:[
        ^ super keyPress:aKey x:x y:y
    ].
    (rawKey = 'Ctrl') ifTrue:[
        ^ self
    ].

"/    Transcript show:'unhandled key: '; showCR:rawKey.

    "Modified: / 5.5.1999 / 18:32:42 / cg"
!

shellTerminated
    "shell has terminated"

"/Delay waitForSeconds:10.
    [self readAnyAvailableData > 0] whileTrue:[Delay waitForSeconds:0.1].
"/
    self closeDownShell.

    shellTerminateAction notNil ifTrue:[
        shellTerminateAction value
    ] ifFalse:[
        "/ may be removed ...
        self warn:'shell terminated'.
    ]

    "Modified: / 5.5.1999 / 18:43:22 / cg"
!

sizeChanged:how
    super sizeChanged:how.
    self defineWindowSize.

    "Modified: / 11.6.1998 / 22:51:48 / cg"
! !

!TerminalView methodsFor:'functions'!

doBackspace
    self cursorLeft.
    self replaceCharAtCursor:(Character space).
    self cursorLeft.

    "Modified: / 10.6.1998 / 17:09:12 / cg"
!

doClearDisplay
    "clear everything"

    self doClearEntireScreen.

    "Modified: / 21.7.1998 / 20:05:35 / cg"
!

doClearEntireLine
    "clear the cursor line. cursor position remains unchanged"

    self at:cursorLine put:''
!

doClearEntireScreen
    "clear everything"

    firstLineShown to:(list size) do:[:l |
        self at:l put:''
    ].

    "Modified: / 21.7.1998 / 20:00:19 / cg"
    "Created: / 21.7.1998 / 20:05:24 / cg"
!

doClearFromBeginningOfLine
    "clear from beginning of line to the cursorPosition"

    |l|

    l := self listAt:cursorLine.
    l notNil ifTrue:[
        (l size >= (cursorCol-1)) ifTrue:[
            l := l copy from:1 to:cursorCol-1 put:(Character space).
        ] ifFalse:[
            l := nil.
        ].
        self withoutRedrawAt:cursorLine put:l.
        self invalidateLine:cursorLine
        "/ self at:cursorLine put:l.
    ]

    "Modified: / 20.6.1998 / 19:10:21 / cg"
    "Created: / 21.7.1998 / 20:10:58 / cg"
!

doClearFromBeginningOfScreen
    "clear from beginning of the screen to the cursorPosition"

    self doClearFromBeginningOfLine.
    cursorLine-1 to:firstLineShown do:[:l |
        self at:l put:''
    ].

    "Modified: / 10.6.1998 / 14:45:43 / cg"
    "Created: / 21.7.1998 / 20:08:29 / cg"
!

doClearToEndOfLine
    "clear from the cursorPosition to the end of the line"

    |l|

    l := self listAt:cursorLine.
    (l size >= (cursorCol-1)) ifTrue:[
        l notNil ifTrue:[
            l := l copyTo:cursorCol-1.
            self withoutRedrawAt:cursorLine put:l.
            self invalidateLine:cursorLine
            "/ self at:cursorLine put:l.
        ]
    ]

    "Created: / 10.6.1998 / 14:45:01 / cg"
    "Modified: / 20.6.1998 / 19:10:21 / cg"
!

doClearToEndOfScreen
    "clear from the cursorPosition to the end of the screen"

    self doClearToEndOfLine.
    cursorLine+1 to:(list size) do:[:l |
        self at:l put:''
    ].

    "Modified: / 10.6.1998 / 14:45:43 / cg"
    "Created: / 21.7.1998 / 20:06:14 / cg"
!

doCursorDown:n
    "move the cursor down by n lines"

    |wasOn rEnd|

"/    rangeEndLine == numberOfLines ifTrue:[
"/        ^ super cursorDown:n
"/    ].
    cursorLine + 1 - firstLineShown + n <= rangeEndLine ifTrue:[
        "/ no special action req'd
        ^ super cursorDown:n
    ].

    n timesRepeat:[
        wasOn := self hideCursor.
        rEnd := rangeEndLine+firstLineShown-1.
        cursorLine ==  rEnd ifTrue:[
            self deleteLine:(rangeStartLine+firstLineShown-1).
            self insertLine:'' before:rEnd.
        ] ifFalse:[
            super cursorDown
        ].
        wasOn ifTrue:[self showCursor]. "/ self makeCursorVisibleAndShowCursor:wasOn.
    ]

    "Modified: / 20.6.1998 / 20:29:39 / cg"
!

doCursorHome
    "move the cursor to the home position"

    self cursorVisibleLine:1 col:1
    "/ super cursorHome

    "Modified: / 10.6.1998 / 20:47:31 / cg"
!

doCursorLeft:n
    "move the cursor to the left by n columns"

    n timesRepeat:[
        super cursorLeft
    ]

    "Created: / 11.6.1998 / 22:30:00 / cg"
!

doCursorNewLine
    "move the cursor down to the next line (col remains unchanged)"

    super cursorDown:1

    "Modified: / 10.6.1998 / 16:55:57 / cg"
!

doCursorReturn
    "move the cursor down and left to the beginning to the next line"

    super cursorToBeginOfLine
!

doCursorRight:n
    "move the cursor to the right by n columns"

    self cursorCol:(cursorCol + n)

    "Created: / 10.6.1998 / 15:10:08 / cg"
!

doCursorUp:n
    "move the cursor up by n lines"

    |wasOn rStart|

"/    rangeStartLine == 1 ifTrue:[
"/        ^ super cursorUp:n
"/    ].
    cursorLine + 1 - firstLineShown - n >= rangeStartLine ifTrue:[
        "/ no special action req'd
        ^ super cursorUp:n
    ].
    n timesRepeat:[
        wasOn := self hideCursor.
        rStart := rangeStartLine+firstLineShown-1.
        cursorLine == rStart ifTrue:[
            (rangeEndLine+firstLineShown-1) <= list size ifTrue:[
                self deleteLine:(rangeEndLine+firstLineShown-1).
                self insertLine:'' before:rStart.
            ].
        ] ifFalse:[
            super cursorUp
        ].
        self makeCursorVisibleAndShowCursor:wasOn.
    ]

    "Created: / 11.6.1998 / 22:29:46 / cg"
    "Modified: / 20.6.1998 / 20:30:34 / cg"
! !

!TerminalView methodsFor:'initialization-shell'!

startCommand:aCommand
    "start a command on a pseudo terminal. If the command arg is nil,
     a shell is started. The command is started in the current directory.
     Also fork a reader process, to read the shells output and
     tell me, whenever something arrives"

    ^ self startCommand:aCommand in:nil

    "Modified: / 20.7.1998 / 18:30:24 / cg"
!

startCommand:aCommand in:aDirectory
    "start a command on a pseudo terminal. If the command arg is nil,
     a shell is started. If aDirectory is not nil, the command is
     executed in that directory.
     Also fork a reader process, to read the shells output and
     tell me, whenever something arrives"

    |p slaveFD execFdArray closeFdArray blocked exitStatus 
     stxToCommandPipe commandToStxPipe cmd shell args env shellAndArgs|

    shellCommand := aCommand.
    shellDirectory := aDirectory.

    self create.  "/ need my windowID (to pass down in environment)

    OperatingSystem isMSWINDOWSlike ifTrue:[
        "/ must use another mechanism

        stxToCommandPipe := OperatingSystem makePipe.
        (stxToCommandPipe isNil
        or:[(stxToCommandPipe at:2) < 0
        or:[(stxToCommandPipe at:2) < 0]]) ifTrue:[
            ^ self warn:'Could not create pipe to COMMAND.COM.'.
        ].

        commandToStxPipe := OperatingSystem makePipe.
        (commandToStxPipe isNil 
        or:[(commandToStxPipe at:2) < 0
        or:[(commandToStxPipe at:2) < 0]]) ifTrue:[
            ^ self warn:'Could not create pipe from COMMAND.COM.'.
        ].

        shellAndArgs := OperatingSystem commandAndArgsForOSCommand:aCommand.
        shell := shellAndArgs at:1.
        args  := (shellAndArgs at:2) ? ''.

        "/ pipe readSide is p at:1;
        "/      writeSide is p at:2

        execFdArray := Array 
                         with:(stxToCommandPipe at:1)
                         with:(commandToStxPipe at:2)
                         with:(commandToStxPipe at:2).

        closeFdArray := #(). 

        outStream := ExternalStream forReadingFromFileDescriptor:(commandToStxPipe at:1).
        inStream := ExternalStream forWritingToFileDescriptor:(stxToCommandPipe at:2).
        outStream buffered:false.
        inStream buffered:false.
    ] ifFalse:[
        p := ExternalStream makePTYPair.
        p isNil ifTrue:[
            self warn:'Cannot open pty.'.
            ^ self.
        ].

        "/ p at:1 is the master;
        "/ p at:2 is the slave
        inStream := outStream := (p at:1).
        inStream buffered:false.

        self defineWindowSize.

        "/ fork a shell process on the slave-side
        slaveFD := (p at:2) fileDescriptor.

        execFdArray := Array with:slaveFD with:slaveFD with:slaveFD.

        "have to close unused filedescriptors.
         Otherwise, listening sockets may be inherited by child processes"

        closeFdArray := 3 to:OperatingSystem maxNumberOfOpenFiles-1. "close anything but stdin, stdout, stderr"
        
        aCommand isNil ifTrue:[
            shell := OperatingSystem getEnvironment:'SHELL'.
            shell size == 0 ifTrue:[
                shell := '/bin/sh'.
            ].
            cmd := shell asFilename baseName.
            args := (Array with:cmd).
        ] ifFalse:[
            shell := '/bin/sh'.
            args := (Array with:'sh' with:'-c' with:aCommand).
        ].
    ].

    blocked := OperatingSystem blockInterrupts.

    shellPid := Processor
               monitor:[
                  OperatingSystem isMSWINDOWSlike ifTrue:[
                      OperatingSystem
                          exec:shell
                          withArguments:args
                          fileDescriptors:execFdArray
                          closeDescriptors:closeFdArray
                          fork:true
                          newPgrp:true
                          inDirectory:aDirectory.
                  ] ifFalse:[
                      env := Dictionary new.
                      env at:'TERM'     put:(self terminalType).
                      env at:'LINES'    put:(numberOfLines printString).
                      env at:'COLUMNS'  put:(numberOfColumns printString).
                      device platformName = 'X11' ifTrue:[  
                          drawableId notNil ifTrue:[
                              env at:'WINDOWID' put:(drawableId address printString).
                          ].
                      ].
                      env at:'SHELL'  put:shell.

                      OperatingSystem
                          exec:shell
                          withArguments:args
                          environment:env
                          fileDescriptors:execFdArray
                          closeDescriptors:closeFdArray
                          fork:true
                          newPgrp:true
                          inDirectory:aDirectory.
                  ]
               ]
               action:[:status |
"/                  Transcript show:'pid:'; showCR:status pid.
"/                  Transcript show:'status:'; showCR:status status.
"/                  Transcript show:'code:'; showCR:status code.
"/                  Transcript show:'core:'; showCR:status core.
                  status stillAlive ifFalse:[
                      exitStatus := status.
                      OperatingSystem closePid:shellPid.
                      shellPid := nil.
                      self pushEvent:#shellTerminated
                  ].
               ].

    blocked ifFalse:[
        OperatingSystem unblockInterrupts
    ].

    shellPid isNil ifTrue:[
        self warn:'Cannot start shell'.
        p notNil ifTrue:[
            (p at:1) close.
            (p at:2) close.
        ]
    ].

    self startReaderProcess.

    "Created: / 20.7.1998 / 18:19:32 / cg"
    "Modified: / 5.5.1999 / 17:28:37 / cg"
!

startShell
    "start a shell on a pseudo terminal in the current directory.
     Also fork a reader process, to read the shells output and
     tell me, whenever something arrives"

    ^ self startCommand:nil

    "
     VT100TerminalView openShell
    "

    "Modified: / 20.7.1998 / 18:29:54 / cg"
!

startShellIn:aDirectory
    "start a shell on a pseudo terminal in some directory.
     Also fork a reader process, to read the shells output and
     tell me, whenever something arrives"

    ^ self startCommand:nil in:aDirectory

    "
     VT100TerminalView openShellIn:'/etc'
    "

    "Modified: / 20.7.1998 / 18:29:46 / cg"
! !

!TerminalView methodsFor:'initialize / release'!

closeDownShell
    "shut down my shell process and stop the background reader thread."

    |pid|

    (pid := shellPid) notNil ifTrue:[
"/        Transcript print:'killing shell pid='; showCR:pid.

        OperatingSystem terminateProcessGroup:pid.
        OperatingSystem terminateProcess:pid.
        Delay waitForSeconds:1.
        shellPid notNil ifTrue:[
            OperatingSystem isMSWINDOWSlike ifFalse:[
                OperatingSystem killProcessGroup:pid.
            ].
            OperatingSystem killProcess:pid.
            shellPid := nil.
        ].
        OperatingSystem closePid:pid.
    ].

    self stopReaderProcess.

    inStream notNil ifTrue:[
        inStream isStream ifTrue:[inStream close].
        inStream := nil
    ].
    outStream notNil ifTrue:[
        outStream close.
        outStream := nil
    ].

    "Modified: / 5.5.1999 / 18:43:02 / cg"
!

destroy
    "destroy myself - shut down the shell, stop the reader thread."

    self closeDownShell.
    self flushInput.
    super destroy

    "Modified: / 5.5.1999 / 18:42:55 / cg"
!

escapeSequences:codes
    "setup my escape sequences"

    |tree|

    tree isNil ifTrue:[tree := escapeSequenceTree := IdentityDictionary new].

    codes do:[:specEntry |
        |sequence function|

        sequence := (specEntry at:1) withEscapes.
        function := specEntry at:2.

        tree := escapeSequenceTree.

        sequence keysAndValuesDo:[:idx :char |
            |followup|

            idx == sequence size ifTrue:[
                tree at:char put:function
            ] ifFalse:[
                followup := tree at:char ifAbsent:nil.
                followup isNil ifTrue:[
                    tree at:char put:(followup := IdentityDictionary new).
                ].
                tree := followup
            ]
        ]
    ].
    escapeLeadingChars := escapeSequenceTree keys asSet.
    escapeLeadingChars add:(Character cr).
    escapeLeadingChars add:(Character return).
    escapeLeadingChars add:(Character backspace).

    escapeLeadingChars := escapeLeadingChars asArray

    "Modified: / 9.6.1998 / 19:43:12 / cg"
!

flushInput
    |sensor|

    "/ flush any leftover input-processing events
    (sensor := self sensor) notNil ifTrue:[
        sensor flushEventsFor:self withType:#processInput:n:.
    ]

    "Modified: / 21.7.1998 / 19:00:13 / cg"
    "Created: / 5.5.1999 / 18:41:42 / cg"
!

initialize
    super initialize.

    showMatchingParenthesis := false.
    insertMode := false.
    alwaysAppendAtEnd := false.
    collectSize := 100.
    st80Mode := false.
    trimBlankLines := true.
    localEcho := false.
    inputTranslateCRToNL := false.

    numberOfColumns := 80.
    numberOfLines := 24.
    rangeStartLine := 1.
    rangeEndLine := numberOfLines.

    self initializeKeyboardSequences.
    list := OrderedCollection new:24 withAll:''.

    self initializeKeyboardMap.

    "
     VT52TerminalView openShell
     VT100TerminalView openShell
    "

    "Modified: / 5.5.1999 / 17:54:47 / cg"
!

initializeKeyboardMap
    |keys ctrlKeys cmdKeys|

    "/ setup my own keyboardMap, where control-keys 
    "/ (and some Cmd-keys) are not translated.
    kbdMap := device keyboardMap copy.
    keys := kbdMap keys.

    ctrlKeys := keys select:[:key | key startsWith:'Ctrl'].
    ctrlKeys do:[:key | kbdMap removeKey:key].

    cmdKeys := keys select:[:key | key startsWith:'Cmd'].
    cmdKeys do:[:key | |val|
        val := kbdMap at:key.
        val isSymbol ifTrue:[
            (#(Copy Paste SaveAs Print) includes:val) ifFalse:[
                kbdMap removeKey:key
            ]
        ]
    ].

    kbdMap removeKey:#Delete ifAbsent:[].
    kbdMap removeKey:#BackSpace ifAbsent:[].

    "
     VT52TerminalView openShell
    "

    "Modified: / 29.4.1999 / 14:25:24 / cg"
!

initializeKeyboardSequences
    kbdSequences := (self anyKeyCodes)

    "Modified: / 5.5.1999 / 15:01:09 / cg"
!

keyboardMap
    "return my keyboard map. This has control keys removed and
     those will be passed unchanged to the shell"

    ^ kbdMap

    "Modified: / 10.6.1998 / 17:46:59 / cg"
!

reinitialize
    "reinit after a snapIn.
     this is invoked (by the system thread) after a snapShot image restart"

    super reinitialize.

    shellPid := nil.

    self stopReaderProcess.
    self flushInput.

    readerProcess := nil.
    inStream := outStream := nil.

    "/ must fork at low-prio (to avoid running reader at prio31)
    [
       "/ Delay waitForSeconds:0.5.
        self doClearEntireScreen.
        self cursorLine:1 col:1.
        self contents:nil.
        self flash.

        "/
        "/ all I can do is to restart the original command
        "/
        self startCommand:shellCommand in:shellDirectory.
    ] forkAt:8

    "Modified: / 5.5.1999 / 18:41:55 / cg"
!

startReaderProcess
    "Start a reader process, which looks for the commands output,
     and sends me #processInput:n: events whenever something arrives."

    readerProcess isNil ifTrue:[
        readerProcess := [
            [
                self waitUntilVisible.

                Stream streamErrorSignal handle:[:ex |
                    Transcript show:'Terminal [error]: '; showCR:ex errorString.
                    Transcript show:'Terminal [info]: '; showCR:OperatingSystem lastErrorString.
                ] do:[
                    [true] whileTrue:[
                        Object abortSignal handle:[:ex |
                            self showCursor.
                        ] do:[
                            |n sensor|

                            outStream readWait.
                            (sensor := self sensor hasKeyPressEventFor:self) ifTrue:[
                                true "(sensor userEventCount > 10)" ifTrue:[
                                    "/ give terminalView a chance to
                                    "/ send out the character.
                                    Delay waitForSeconds:0.01.
                                ]
                            ] ifFalse:[
                                n := self readAnyAvailableData.
                                n > 0 ifTrue:[
                                    shellPid notNil ifTrue:[
                                        self waitForOutputToDrain.
                                    ]
                                ] ifFalse:[
                                    n == 0 ifTrue:[
                                        "/ Windows IPC has a bug - it always
                                        "/ returns 0 (when the command is idle)
                                        "/ and says its at the end (sigh)

                                        OperatingSystem isMSWINDOWSlike ifTrue:[
                                            Delay waitForSeconds:0.01
                                        ] ifFalse:[
                                            outStream atEnd ifTrue:[
                                                outStream close. outStream := nil.
                                                inStream close.  inStream := nil.
                                                
                                                Processor activeProcess terminate.
                                            ] ifFalse:[
                                                "/ this should not happen.

                                                Delay waitForSeconds:0.05
                                            ]
                                        ].
                                    ]
                                ]
                            ]
                        ]
                    ]
                ]
            ] valueOnUnwindDo:[
                readerProcess := nil    
            ]
        ] fork. "/ forkAt:9.
        readerProcess name:'pty reader'.
    ]

    "
     VT100TerminalView openShell
    "

    "Modified: / 5.5.1999 / 17:58:02 / cg"
    "Modified: / 28.1.2002 / 21:10:13 / micha"
!

stopReaderProcess
    "stop the background reader thread"

    |sensor|

    readerProcess notNil ifTrue:[
        readerProcess terminate.
        "/ give it a chance to really terminate
        Processor yield.
        readerProcess := nil
    ].

    "/ flush any leftover input-processing events
    (sensor := self sensor) notNil ifTrue:[
        sensor flushEventsFor:self withType:#processInput:n:.
    ]
    "Modified: / 21.7.1998 / 19:00:13 / cg"
! !

!TerminalView methodsFor:'menu'!

doReset
    "reset the scroll-range; 
     may have to reset more in the future (current font-set; color; etc)"

    rangeStartLine := 1.
    rangeEndLine := numberOfLines.

    self normal.

!

doSendInterrupt
    "send an INT-signal to the shell (UNIX only)"

    OperatingSystem sendSignal:(OperatingSystem sigINT) to:shellPid negated.

    "Modified: / 10.6.1998 / 17:49:49 / cg"
!

editMenu
    "return the views middleButtonMenu"

    <resource: #keyboard (#Copy #Paste #Print #SaveAs #Print)>
    <resource: #programMenu>

    |items m sensor|

    ((sensor := self sensor) notNil and:[sensor ctrlDown]) ifTrue:[
        items := #(
                        ('Interrupt'      doSendInterrupt)  
                        ('-'                                    )
                        ('Reset'          doReset)  
                  ).
    ] ifFalse:[
        items := #(
                        ('Copy'         copySelection    Copy   )
                        ('Paste'        pasteOrReplace   Paste  )
                        ('-'                                    )
                        ('Font...'      changeFont              )
                        ('-'                                    )
                        ('Save as...'   save             SaveAs )
                        ('Print'        doPrint          Print  )
                  ).

"/ disabled - for now;
"/ need a more intelligent filtering of control characters.
"/
"/        filterStream isNil ifTrue:[
"/           items := items , #(
"/                        ('-'                                    )
"/                        ('Start Save as ...'  startSaveAs       )
"/                        ('Start Print'        startPrint        )
"/                             )
"/        ] ifFalse:[
"/           items := items , #(
"/                        ('-'                                    )
"/                        ('Stop filter'  stopFilter              )
"/                             )
"/        ].
    ].

    m := PopUpMenu itemList:items resources:resources.

    self hasSelection not ifTrue:[
        m disable:#copySelection.
    ].
    ^ m.

    "Modified: / 29.4.1999 / 11:12:56 / cg"
!

startSaveAs
    "start saving all received data to some file"

    |fn|

    fn := Dialog requestFileName:'Save received data in file:'.
    fn size > 0 ifTrue:[
        filterStream := fn asFilename writeStream
    ].

    "Created: / 29.4.1999 / 11:06:29 / cg"
!

stopFilter
    "stop saving/printing of received data"

    filterStream close.
    filterStream := nil.

    "Created: / 29.4.1999 / 11:07:49 / cg"
    "Modified: / 29.4.1999 / 11:09:52 / cg"
! !

!TerminalView methodsFor:'misc'!

removeTrailingBlankLines
    ^ self
! !

!TerminalView methodsFor:'processing - input'!

doNothing
    "private - end of an ignored escape-sequence"

    self endOfSequence

    "Created: / 12.6.1998 / 20:40:43 / cg"
!

endOfSequence
    "private - reset state-machine at end of escape-sequence"

    state := 0.

    "Created: / 12.6.1998 / 20:39:52 / cg"
!

nextPutAll:aString
    "/ self processInput:aString n:aString size
    self pushEvent:#processInput:n: with:aString with:aString size.

    "Created: / 27.7.1998 / 15:10:59 / cg"
    "Modified: / 27.7.1998 / 23:16:19 / cg"
!

processInput:buffer n:count
    |sensor|

    filterStream notNil ifTrue:[
        filterStream nextPutAll:(buffer copyTo:count).
    ].

"/    self hideCursor.

    "/ the following may not be too clean, but adds a lot of speed.
    "/ instead of passing every individual character through the
    "/ escape-state machine, collect chunks of non-control text
    "/ when in state 0, and add them immediately to the pendingLines
    "/ collection of the textCollectors asynchronous update mechanism.
    "/ This helps a lot if you do something like "ls -lR /" ...
    "/ For debugging the state machine, reenable the commented lines
    "/ below.

"/1 to:count do:[:i|
"/    self nextPut:(buffer at:i).
"/].
"/self showCursor.
"/^ self.

    access critical:[
        |i i2 s crnlFollows|

        i := 1.
        [i <= count] whileTrue:[
            (state == 0) ifTrue:[
                "/ in initial state.
                "/ quick scan forward for next control character ...
                i2 := buffer indexOfControlCharacterStartingAt:i.
                i2 > count ifTrue:[i2 := 0].

                i2 == 0 ifTrue:[
                    "/ no control characters - simply append all
                    "/ to the outstanding lines ...
                    s := buffer copyFrom:i to:count.
                    i := count + 1. "/ leave loop.
                    crnlFollows := false.
                ] ifFalse:[
                    i2 > i ifTrue:[
                        s := buffer copyFrom:i to:i2-1.
                        i := i2. "/ proceed withcontrol character
                        crnlFollows := false.
                        i < (count - 1) ifTrue:[
                            (buffer at:i) == Character return ifTrue:[
                                (buffer at:i+1) == Character nl ifTrue:[
                                    crnlFollows := true.
                                    i := i + 2.
                                ]
                            ]
                        ].
                    ]
                ].
            ].

            s notNil ifTrue:[
                currentEmphasis notNil ifTrue:[
                    s := s emphasizeAllWith:currentEmphasis
                ].

                outstandingLine size > 0 ifTrue:[
                    outstandingLine := outstandingLine , s.
                ] ifFalse:[
                    outstandingLine := s.
                ].
                crnlFollows ifTrue:[
                    outstandingLines isNil ifTrue:[
                        outstandingLines := OrderedCollection with:outstandingLine
                    ] ifFalse:[
                        outstandingLines add:outstandingLine.
                    ].
                    outstandingLine := ''.
                ].
                s := nil.

                collecting ifTrue:[
                    flushPending ifFalse:[
                        self installDelayedUpdate
                    ] ifTrue:[
    "/                    outstandingLines size > collectSize ifTrue:[
    "/                        self endEntry
    "/                    ]
                    ]
                ] ifFalse:[
                    self endEntry
                ].
            ] ifFalse:[
                "/ no chunk to append (in an escape sequence)
                "/ must handle individual characters
                "/ to update the state machine.

                self nextPut:(buffer at:i).
                i := i + 1.
                [state ~~ 0 and:[i <= count]] whileTrue:[
                    self nextPut:(buffer at:i).
                    i := i + 1.
                ]
            ]
        ].
    ].

    state == 0 ifTrue:[
        (sensor := self sensor) notNil ifTrue:[
            "/ if there is no more output pending from the shell,
            "/ enforce update of the view (asynchronous)

            (sensor hasEvent:#processInput:n: for:self) ifFalse:[
                self endEntry.
                self showCursor.
                "/ self makeCursorVisible.
            ] ifTrue:[
                "/ if there is more output pending from the shell,
                "/ and many lines have already been collected,
                "/ also enforce update of the view (asynchronous)
                "/ Thus, it will update at least once for every
                "/ collectSize lines.

                outstandingLines size > collectSize ifTrue:[ 
                    self endEntry.
                    self showCursor.

                    "/ make certain that things are really displayed ...
                    windowGroup notNil ifTrue:[
                        windowGroup processRealExposeEventsFor:self.
                    ]
                ].
            ]
        ].
    ].

    "Created: / 10.6.1998 / 17:26:09 / cg"
    "Modified: / 29.4.1999 / 11:08:48 / cg"
    "Modified: / 28.1.2002 / 20:41:36 / micha"
!

readAnyAvailableData
    "read data from the stream,
     and sends me #processInput:n: events if something arrived.
     Returns the amount of data read."

    |buffer n bufferSize|

    outStream isNil ifTrue:[^ 0].

    bufferSize := 512. "/ 1024.

    buffer := String new:bufferSize.
    ExternalStream readErrorSignal handle:[:ex |
        n := 0
    ] do:[
        n := outStream nextAvailableBytes:bufferSize into:buffer startingAt:1.
    ].
    n > 0 ifTrue:[
        self pushEvent:#processInput:n: with:buffer with:n.
    ].

    ^ n

    "Created: / 5.5.1999 / 17:57:39 / cg"
    "Modified: / 8.5.1999 / 20:14:14 / cg"
    "Modified: / 28.1.2002 / 21:10:25 / micha"
!

sync
    self waitForOutputToDrain

    "Created: / 27.7.1998 / 23:49:44 / cg"
!

waitForOutputToDrain
    |sensor|

    sensor := self sensor.
    (sensor notNil and:[sensor userEventCount > 30]) ifTrue:[
        [sensor userEventCount > 5] whileTrue:[
            "/ give terminalView a chance to
            "/ catch up.
            Delay waitForSeconds:0.1.
        ]
    ].

    "Created: / 27.7.1998 / 23:47:22 / cg"
    "Modified: / 5.5.1999 / 18:51:00 / cg"
! !

!TerminalView methodsFor:'queries'!

preferredExtent
    "return my preferred extent - this is computed from my numberOfLines,
     numberOfCols and font size"

    ^ (fontWidth * self class defaultNumberOfColumns + (leftMargin * 2))
      @ 
      ((self heightForLines:self class defaultNumberOfLines) + 8)

    "Modified: / 20.6.1998 / 20:06:57 / cg"
!

terminalType
    "returns a string describing this terminal (usually, this is
     passed down to the shell as TERM environment variable).
     Here, 'dump' is returned."

    ^ 'dump'

    "Created: / 10.6.1998 / 16:22:30 / cg"
    "Modified: / 5.5.1999 / 11:22:32 / cg"
! !

!TerminalView methodsFor:'selection handling'!

autoMoveCursorToEndOfSelection
    "Redefined to return false since the cursor should
     not be affected by selecting"

    ^ false

!

paste:someText
    "paste - redefined to send the chars to the shell instead
     of pasting into the view"

    |s nLines|

    s := someText.
    s isString ifTrue:[
        s := s asStringCollection
    ] ifFalse:[
        (s isKindOf:StringCollection) ifFalse:[
            self warn:'selection (' , s class name , ') is not convertable to Text'.
            ^ self
        ]
    ].
    (nLines := s size) == 0 ifTrue:[^ self].
    (nLines == 1 and:[(s at:1) size == 0]) ifTrue:[^ self].
    s keysAndValuesDo:[:idx :line |
        line notNil ifTrue:[inStream nextPutAll:line].
        idx ~~ nLines ifTrue:[
            inStream nextPut:(Character return).
        ]
    ].

    "Modified: / 12.6.1998 / 22:12:47 / cg"
! !

!TerminalView class methodsFor:'documentation'!

version
    ^ '$Header: /cvs/stx/stx/libwidg2/TerminalView.st,v 1.94 2002-06-17 13:09:32 stefan Exp $'
! !
